/*
 * System call handlers
 * Copyright (c) 2003, Jeffrey K. Hollingsworth <hollings@cs.umd.edu>
 * Copyright (c) 2003,2004 David Hovemeyer <daveho@cs.umd.edu>
 * $Revision: 1.59 $
 * 
 * This is free software.  You are permitted to use,
 * redistribute, and modify it as specified in the file "COPYING".
 */

#include <geekos/syscall.h>
#include <geekos/errno.h>
#include <geekos/kthread.h>
#include <geekos/int.h>
#include <geekos/elf.h>
#include <geekos/malloc.h>
#include <geekos/keyboard.h>
#include <geekos/string.h>
#include <geekos/user.h>
#include <geekos/timer.h>
#include <geekos/vfs.h>

struct semaphore** sem = NULL;
extern int policy;
extern int g_Quantum;

/*
 * Null system call.
 * Does nothing except immediately return control back
 * to the interrupted user program.
 * Params:
 *  state - processor registers from user mode
 *
 * Returns:
 *   always returns the value 0 (zero)
 */
static int Sys_Null(struct Interrupt_State* state)
{
	return 0;
}

/*
 * Exit system call.
 * The interrupted user process is terminated.
 * Params:
 *   state->ebx - process exit code
 * Returns:
 *   Never returns to user mode!
 */
static int Sys_Exit(struct Interrupt_State* state)
{
	struct User_Context* pUserContext = g_currentThread->userContext;

	Free_Segment_Descriptor(pUserContext->ldtDescriptor);
	Free(pUserContext->memory);
	Free(pUserContext);

	Exit(state->ebx);
}

/*
 * Print a string to the console.
 * Params:
 *   state->ebx - user pointer of string to be printed
 *   state->ecx - number of characters to print
 * Returns: 0 if successful, -1 if not
 */
static int Sys_PrintString(struct Interrupt_State* state)
{
	char* msg = 0;
	int i = (int)state->ecx;

	msg = (char*)Malloc(i+1);

	if(msg == 0)
		return -1;

	if(!Copy_From_User(msg, state->ebx, i+1))
	{
		Free(msg);
		return -1;
	}

	Print(msg);

	return 0;
}

/*
 * Get a single key press from the console.
 * Suspends the user process until a key press is available.
 * Params:
 *   state - processor registers from user mode
 * Returns: the key code
 */
static int Sys_GetKey(struct Interrupt_State* state)
{
	Keycode key = Wait_For_Key();

	return key;
}

/*
 * Set the current text attributes.
 * Params:
 *   state->ebx - character attributes to use
 * Returns: always returns 0
 */
static int Sys_SetAttr(struct Interrupt_State* state)
{
	TODO("SetAttr system call");
}

/*
 * Get the current cursor position.
 * Params:
 *   state->ebx - pointer to user int where row value should be stored
 *   state->ecx - pointer to user int where column value should be stored
 * Returns: 0 if successful, -1 otherwise
 */
static int Sys_GetCursor(struct Interrupt_State* state)
{
	int row, col;

	Get_Cursor(&row, &col);
	Copy_To_User(state->ebx, &row, sizeof(int));
	Copy_To_User(state->ecx, &col, sizeof(int));

	return 0;
}

/*
 * Set the current cursor position.
 * Params:
 *   state->ebx - new row value
 *   state->ecx - new column value
 * Returns: 0 if successful, -1 otherwise
 */
static int Sys_PutCursor(struct Interrupt_State* state)
{
	Put_Cursor(state->ebx, state->ecx);
	return 0;
}

/*
 * Create a new user process.
 * Params:
 *   state->ebx - user address of name of executable
 *   state->ecx - length of executable name
 *   state->edx - user address of command string
 *   state->esi - length of command string
 * Returns: pid of process if successful, error code (< 0) otherwise
 */
static int Sys_Spawn(struct Interrupt_State* state)
{
	char* path;
	char* command;
	struct Kernel_Thread* pthread = 0;
	int pid;

	path = (char*)Malloc(state->ecx+1);
	command = (char*)Malloc(state->esi+1);

	Copy_From_User((void*)path, state->ebx, state->ecx+1);
	Copy_From_User((void*)command, state->edx, state->esi+1);

	Enable_Interrupts();
	pid = Spawn(path, command, &pthread);
	Disable_Interrupts();

	Free(path);
	Free(command);

	if(pid < 1 )
		return -1;
	return pid;

}

/*
 * Wait for a process to exit.
 * Params:
 *   state->ebx - pid of process to wait for
 * Returns: the exit code of the process,
 *   or error code (< 0) on error
 */
static int Sys_Wait(struct Interrupt_State* state)
{
	int exitcode;
	struct Kernel_Thread* pthread = Lookup_Thread(state->ebx);

	Enable_Interrupts();
	exitcode = Join(pthread);
	Disable_Interrupts();

	return exitcode;
}

/*
 * Get pid (process id) of current thread.
 * Params:
 *   state - processor registers from user mode
 * Returns: the pid of the current thread
 */
static int Sys_GetPID(struct Interrupt_State* state)
{
	TODO("GetPID system call");
}

/*
 * Set the scheduling policy.
 * Params:
 *   state->ebx - policy,
 *   state->ecx - number of ticks in quantum
 * Returns: 0 if successful, -1 otherwise
 */
static int Sys_SetSchedulingPolicy(struct Interrupt_State* state)
{
	int i;
	struct Kernel_Thread* pthread;

	policy = state->ebx;
	g_Quantum = state->ecx;

	if(g_Quantum < 1)
		return -1;

	return 0;
}

/*
 * Get the time of day.
 * Params:
 *   state - processor registers from user mode
 *
 * Returns: value of the g_numTicks global variable
 */
static int Sys_GetTimeOfDay(struct Interrupt_State* state)
{
	return g_numTicks;
}

/*
 * Create a semaphore.
 * Params:
 *   state->ebx - user address of name of semaphore
 *   state->ecx - length of semaphore name
 *   state->edx - initial semaphore count
 * Returns: the global semaphore id
 */
static int Sys_CreateSemaphore(struct Interrupt_State* state)
{
	//Project4
}

/*
 * Acquire a semaphore.
 * Assume that the process has permission to access the semaphore,
 * the call will block until the semaphore count is >= 0.
 * Params:
 *   state->ebx - the semaphore id
 *
 * Returns: 0 if successful, error code (< 0) if unsuccessful
 */
static int Sys_P(struct Interrupt_State* state)
{
	//Project4
}

/*
 * Release a semaphore.
 * Params:
 *   state->ebx - the semaphore id
 *
 * Returns: 0 if successful, error code (< 0) if unsuccessful
 */
static int Sys_V(struct Interrupt_State* state)
{
	//Project4
}

/*
 * Destroy a semaphore.
 * Params:
 *   state->ebx - the semaphore id
 *
 * Returns: 0 if successful, error code (< 0) if unsuccessful
 */
static int Sys_DestroySemaphore(struct Interrupt_State* state)
{
	//Project4
}

/*
 * Global table of system call handler functions.
 */
const Syscall g_syscallTable[] = {
	Sys_Null,
	Sys_Exit,
	Sys_PrintString,
	Sys_GetKey,
	Sys_SetAttr,
	Sys_GetCursor,
	Sys_PutCursor,
	Sys_Spawn,
	Sys_Wait,
	Sys_GetPID,
	/* Scheduling and semaphore system calls. */
	Sys_SetSchedulingPolicy,
	Sys_GetTimeOfDay,
	Sys_CreateSemaphore,
	Sys_P,
	Sys_V,
	Sys_DestroySemaphore,
};

/*
 * Number of system calls implemented.
 */
const int g_numSyscalls = sizeof(g_syscallTable) / sizeof(Syscall);
